package de.mdelab.mlsdm.interpreter.coverage;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;

import de.mdelab.mlcallactions.CallAction;
import de.mdelab.mlcallactions.CallActionParameter;
import de.mdelab.mlcore.MLElementWithUUID;
import de.mdelab.mlsdm.Activity;
import de.mdelab.sdm.interpreter.core.debug.SDCoverageNotificationReceiver;
import de.mdelab.sdm.interpreter.core.notifications.ActivityExecutionStartedNotification;
import de.mdelab.sdm.interpreter.core.notifications.InterpreterNotification;

public class MLSDMCoverageNotificationReceiver extends SDCoverageNotificationReceiver<CoverageReport, EClassifier> {

	private final HashMap<MLElementWithUUID, CoverageEntry>	coverageEntryMap;

	private final Map<Activity, ActivityCoverageReport>		activityCoverageReports;

	public MLSDMCoverageNotificationReceiver() {
		super();

		this.coverageEntryMap = new HashMap<MLElementWithUUID, CoverageEntry>();

		this.activityCoverageReports = new HashMap<Activity, ActivityCoverageReport>();
	}

	@Override
	public void notifyChanged(final InterpreterNotification<EClassifier> notification)
	{
		if (!(notification.getMainStoryDiagramElement() instanceof MLElementWithUUID))
		{
			return;
		}

		switch (notification.getNotificationType())
		{
			case ACTIVITY_EXECUTION_STARTED:
			{
				/*
				 * make sure that there is an entry for the activity and all its
				 * elements
				 */
				@SuppressWarnings("unchecked")
				final Activity activity = ((ActivityExecutionStartedNotification<Activity, EClassifier>) notification).getActivity();

				ActivityCoverageReport activityCoverageReport = this.activityCoverageReports.get(activity);

				if (activityCoverageReport == null)
				{
					/*
					 * Create an activity coverage report for the activity
					 */
					activityCoverageReport = CoverageFactory.eINSTANCE.createActivityCoverageReport();
					activityCoverageReport.setActivity(activity);

					this.activityCoverageReports.put(activity, activityCoverageReport);

					this.coverageReport.getActivityCoverageReports().add(activityCoverageReport);

					/*
					 * Create coverage entries for the activity's contents
					 */
					final TreeIterator<EObject> it = activity.eAllContents();

					while (it.hasNext())
					{
						final EObject eObject = it.next();

						assert eObject instanceof MLElementWithUUID;

						if (eObject instanceof CallAction || eObject instanceof CallActionParameter)
						{
							/*
							 * Skip call actions
							 */
						}
						else
						{
							final CoverageEntry entry = CoverageFactory.eINSTANCE.createCoverageEntry();
							entry.setElement((MLElementWithUUID) eObject);

							this.coverageEntryMap.put((MLElementWithUUID) eObject, entry);

							activityCoverageReport.getCoverageEntries().add(entry);
						}
					}
				}

				activityCoverageReport.setExecutions(activityCoverageReport.getExecutions() + 1);
				break;
			}

			case ACTIVITY_NODE_EXECUTION_STARTED:
			case EVALUATING_EXPRESSION:
			case INSTANCE_LINK_CREATED:
			case INSTANCE_LINK_DESTROYED:
			case INSTANCE_OBJECT_CREATED:
			case INSTANCE_OBJECT_DESTROYED:
			case STORY_PATTERN_OBJECT_BOUND:
			case STORY_PATTERN_OBJECT_NOT_BOUND:
			case STORY_PATTERN_OBJECT_BINDING_REVOKED:
			case TRAVERSING_ACTIVITY_EDGE:
			case TRAVERSING_LINK:
			case LINK_CHECK_FAILED:
			case LINK_CHECK_SUCCESSFUL:
			{
				final CoverageEntry entry = this.coverageEntryMap.get(notification.getMainStoryDiagramElement());

				assert entry != null;

				entry.setExecutions(entry.getExecutions() + 1);
			}

			default:
				break;
		}
	}

	@Override
	protected CoverageReport createCoverageReport() {
		return CoverageFactory.eINSTANCE.createCoverageReport();
	}
}
