package de.mdelab.mlsdm.interpreter.eclipse;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.MLSDMInterpreter;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMNotificationEmitter;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.eclipse.EclipseExpressionInterpreterManager;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;

/**
 * Implementation of the SDM interpreter for the MLSDM metamodel and use in
 * Eclipse.
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMEclipseInterpreter extends MLSDMInterpreter {

	public MLSDMEclipseInterpreter(	final ClassLoader classLoader,
									final NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> notificationEmitter,
									final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_facadeFactory )
	throws SDMException	{
		super(	new EclipseExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>( MLSDMMetamodelFacadeFactory.INSTANCE.getExpressionFacade(), classLoader, notificationEmitter),
				notificationEmitter,
				p_facadeFactory );
	}

	public MLSDMEclipseInterpreter(final ClassLoader classLoader)
	throws SDMException	{
		this( classLoader, new MLSDMNotificationEmitter(), MLSDMMetamodelFacadeFactory.INSTANCE );
	}

	public MLSDMEclipseInterpreter(	final ClassLoader classLoader,
									final MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_facadeFactory )
	throws SDMException	{
		this( classLoader, new MLSDMNotificationEmitter(), p_facadeFactory );
	}
}
