package de.mdelab.mlsdm.interpreter.facade;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.sdm.interpreter.core.facade.IInstanceFacade;

public class OptimizedMLSDMMetamodelFacadeFactory extends MLSDMMetamodelFacadeFactory {
	
	private final Set<EPackage> metamodels;
	
	private final OptimizedMSLDMInstanceFacade instanceFacade;
	
	public OptimizedMLSDMMetamodelFacadeFactory( final Set<EPackage> p_packages ) {
		metamodels = p_packages;

		instanceFacade = new OptimizedMSLDMInstanceFacade( getUnidirectionalReferences( metamodels ) );
	}

	@Override
	protected IInstanceFacade<AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createInstanceFacade() {
		return getInstanceFacade();
	}

	@Override
	public OptimizedMSLDMInstanceFacade getInstanceFacade() {
		return instanceFacade;
	}
	
	public static Set<EReference> getUnidirectionalReferences( final Set<EPackage> p_packages ) {
		final  Set<EReference> refs = new HashSet<EReference>();

		for ( final EPackage ePackage : p_packages ) {
			refs.addAll( getUnidirectionalReferences( ePackage ) );
		}
		
		return refs;
	}
	
	public static Set<EReference> getUnidirectionalReferences( final EPackage p_package ) {
		final  Set<EReference> refs = new HashSet<EReference>();
		
		for ( final EClassifier classifier : p_package.getEClassifiers() ) {
			if ( classifier instanceof EClass ) {
				for ( final EReference eReference : ( (EClass) classifier ).getEReferences() ) {
					if ( !eReference.isContainment() && eReference.getEOpposite() == null )	{
						refs.add( eReference );
					}
				}
			}
		}

		for ( final EPackage ePackage : p_package.getESubpackages() ) {
			refs.addAll( getUnidirectionalReferences( ePackage ) );
		}
		
		return refs;
	}
}
