package de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreterManager;
import de.mdelab.sdm.interpreter.core.facade.MetamodelFacadeFactory;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.DefaultMatchingStrategy;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.ECheckResult;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.EMatchType;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchStoryPatternObjectTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPart;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPartBasedMatcher;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMNACPatternPart extends PatternPart<AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>
{

	private final StoryPattern					nacPattern;

	private final AbstractStoryPatternObject[]	storyPatternObjects;

	public MLSDMNACPatternPart(
			final PatternPartBasedMatcher<?, ?, ?, ?, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, ?, MLExpression> patternMatcher,
			final StoryPattern nacPattern)
	{
		super(patternMatcher);

		assert nacPattern != null;
		assert nacPattern.getContainingPattern() != null;

		this.nacPattern = nacPattern;

		
		/*
		 * Collect story pattern objects connected to this NAC pattern
		 */
		final Set<AbstractStoryPatternObject> spos = new HashSet<AbstractStoryPatternObject>();

		/*
		 * MB: this conflicts with how the checking of NACPatternParts is handled:
		 * The check is only _performed_ after all objects in the containing pattern
		 * have been bound. However, the check is only _triggered_ by the matcher when
		 * a new binding for an object returned by getStoryPatternObjects() is created.
		 * This means that if the last spo matched by the matcher is not connected to
		 * the nac, the nac is actually never checked.
		 * Since the PatternPartBasedMatcher aims to enforce isomorphism of the nac-
		 * match with the parent pattern match, the nac _has_ to be checked last.
		 * Therefore, it needs dependencies to all spos in the parent pattern.
		 * Problems may occur in conjunction with optional spos.
		 */
//		for (final AbstractStoryPatternLink link : nacPattern.getStoryPatternLinks())
//		{
//			if (link.getSource().getStoryPattern() == nacPattern.getContainingPattern())
//			{
//				spos.add(link.getSource());
//
//				assert link.getTarget().getStoryPattern() == nacPattern || link instanceof MapEntryLink		//TODO is this constraint required at all?
//						&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern;
//			}
//			if (link.getTarget().getStoryPattern() == nacPattern.getContainingPattern())
//			{
//				spos.add(link.getTarget());
//
//				assert link.getSource().getStoryPattern() == nacPattern || link instanceof MapEntryLink		//TODO is this constraint required at all?
//						&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern;
//			}
//
//			if (link instanceof MapEntryLink
//					&& ((MapEntryLink) link).getValueTarget().getStoryPattern() == nacPattern.getContainingPattern())
//			{
//				spos.add(((MapEntryLink) link).getValueTarget());
//			}
//		}
		for(final AbstractStoryPatternObject spo:nacPattern.getContainingPattern().getStoryPatternObjects()) {
			if(!patternMatcher.getFacadeFactory().getStoryPatternObjectFacade().isCreate(spo)) {
				spos.add(spo);
			}
		}

		this.storyPatternObjects = spos.toArray(new AbstractStoryPatternObject[spos.size()]);
	}

	@Override
	public AbstractStoryPatternObject[] getStoryPatternObjects()
	{
		return this.storyPatternObjects;
	}

	@Override
	public EMatchType getMatchType()
	{
		return EMatchType.MANDATORY;
	}

	@Override
	public ECheckResult check()
	throws SDMException	{
		/*
		 * If all story pattern objects are bound
		 */
		for ( final AbstractStoryPatternObject spo : nacPattern.getContainingPattern().getStoryPatternObjects() ) {
			if ( !patternMatcher.isBound( spo ) && !patternMatcher.getFacadeFactory().getStoryPatternObjectFacade().isCreate( spo ) ) {
				return ECheckResult.UNKNOWN;
			}
		}

		/*
		 * Try to find a match for all contained pattern parts
		 */

		/*
		 * Create a new pattern matcher for the sub pattern
		 */
		try	{
			@SuppressWarnings("unchecked")
			final MLSDMPatternPartBasedMatcher subPatternMatcher = new MLSDMPatternPartBasedMatcher(
					this.nacPattern,
					(VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getVariablesScope(),
					/*
					 * MB: the parent matcher's strategy may have an internal state
					 * (e.g. if it's a LogReproducingMatchingStrategy), which will not
					 * be suitable for matching the NAC pattern or may be changed
					 * during the NAC pattern's matching.
					 */
//					(MatchingStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
//							.getMatchingStrategy(),
					new DefaultMatchingStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>((MetamodelFacadeFactory<?, ?, ?, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) patternMatcher.getFacadeFactory()),
					(MetamodelFacadeFactory<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getFacadeFactory(),
					(ExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getExpressionInterpreterManager(),
					(NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.patternMatcher
							.getNotificationEmitter()) {
				@Override
				protected boolean analyzeStoryPatternObjects(boolean secondRun)
				throws SDMException	{
					/*
					 * analyze all story pattern objects
					 */
					for ( final AbstractStoryPatternObject spo : new HashSet<AbstractStoryPatternObject>( this.unboundSPO ) ) {
						/*
						 * All story pattern objects outside of the NAC pattern
						 * are treated as bound objects.
						 */
						final boolean bound = !MLSDMNACPatternPart.this.nacPattern.getStoryPatternObjects().contains(spo);

						if (bound)
						{
							final Variable<EClassifier> variable = this.getVariablesScope().getVariable(spo.getName());

							if (variable == null)
							{
								throw new UnsupportedOperationException();
							}

							/*
							 * Check that the type of the existing variable
							 * matches the type of the story pattern object,
							 * check isomorphism condition and object
							 * constraints
							 */
							if (variable.getValue() != null
									&& this.checkTypeConformance(this.spoFacade.getClassifier(spo), variable.getValue())
									&& this.checkIsomorphism(variable.getValue())
									&& this.checkStoryPatternObjectConstraints(spo, variable.getValue()))
							{
								/*
								 * push matchSpoTransaction onto stack
								 */
								final MatchStoryPatternObjectTransaction<AbstractStoryPatternObject> matchStoryPatternObjectTransaction = new MatchStoryPatternObjectTransaction<AbstractStoryPatternObject>(
										spo, variable.getValue(), this.unboundSPO, this.boundSPO, this.boundInstanceObjects);
								matchStoryPatternObjectTransaction.commit();
								this.matchingStack.push(matchStoryPatternObjectTransaction);

								this.getNotificationEmitter().storyPatternObjectBound(spo, variable.getValue(), this.getVariablesScope(),
										this);
							}
							else
							{
								throw new UnsupportedOperationException();
							}
						}
						else
						{
							/*
							 * Delete existing variables of unbound story
							 * pattern objects
							 */
							this.getVariablesScope().deleteVariable(this.spoFacade.getName(spo));
						}
					}

					return true;
				}
			};

			MATCHLOOP: while ( subPatternMatcher.findNextMatch() ) {
				/*
				 * Check isomorphism
				 */
				StoryPattern pattern = this.nacPattern;

				/*
				 * Collect bound objects of this pattern and its parent
				 * patterns.
				 */
				final Set<Object> boundObjects = new HashSet<Object>();

				do {
					for ( final AbstractStoryPatternObject spo : pattern.getStoryPatternObjects() ) {
						if(subPatternMatcher.getFacadeFactory().getStoryPatternObjectFacade().isCreate(spo)) {
							continue;
						}
						
						final Variable<EClassifier> var = subPatternMatcher.getVariablesScope().getVariable( spo.getName() );

						assert var != null;
						assert var.getValue() != null;

						if ( boundObjects.contains( var.getValue() ) ) {
							continue MATCHLOOP;
						}
						else {
							boundObjects.add( var.getValue() );
						}
					}

					pattern = pattern.getContainingPattern();
				}
				while ( pattern != null );

				return ECheckResult.FAIL;
			}

			/*
			 * No match for the NAC found
			 */
			return ECheckResult.OK;
		}
		catch (final Exception ex)
		{
			throw new RuntimeException(ex);
		}

	}

	@Override
	public boolean match(final MatchState matchState) throws SDMException
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public void createObjects() throws SDMException
	{
		// Do nothing
	}

	@Override
	public void createLinks() throws SDMException
	{
		// Do nothing
	}

	@Override
	public void destroyObjects()
	{
		// Do nothing
	}

	@Override
	public int calculateMatchingCost()
	{
		return PatternPart.MATCHING_NOT_POSSIBLE;
	}

	@Override
	public MatchState createMatchState()
	{
		return null;
	}

	@Override
	public void destroyLinks(final Map<AbstractStoryPatternObject, Object> deletedObjects)
	{
		// Do nothing
	}

}
