package de.mdelab.mlsdm.interpreter.searchModel.model.action;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBasedMatcher;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;

public class MLSDMIndexStepMatchingAction extends MLSDMIndexMatchingAction {

	public MLSDMIndexStepMatchingAction(int id,
			MLSDMSearchModelBasedMatcher matcher, ICLConstraint iclConstraint, MLStringExpression expression,
			Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode,
			Collection<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requirements) {
		super(id, matcher, iclConstraint, expression, iclValueToPatternNode);
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> requirement:requirements) {
			this.addRequiredBinding(requirement);
		}
		this.iclConstraint = iclConstraint;
		this.expression = expression;
		this.index = (Index) matcher.getVariable(iclConstraint.getIndexID()).getValue();
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction#doEstimateCardinality()
	 */
	@Override
	protected double doEstimateCardinality() {
		boolean checkOnly = true;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dependency:getPatternConstraint().getDependencies()) {
			boolean isRequired = getRequiredBindings().contains(dependency);
			if((isRequired && !dependency.isBound()) || (!isRequired && dependency.isBound())) {
				return PatternConstraint.MATCHING_NOT_POSSIBLE;
			}
			
			if(!dependency.isBound()) {
				checkOnly = false;
			}
		}
		if(checkOnly) {
			return PatternConstraint.CHECK_ONLY;
		}
		
		List<Object> query = constructQuery();
		int cardinality = index.estimateCardinality(query);
		
		/*
		 * If the query has no result for sure, this can be returned to allow early pruning.
		 * Otherwise only return the cardinality if the query is actually executable.
		 */
		return (cardinality == 0 || isExecutable()) ? cardinality : PatternConstraint.MATCHING_NOT_POSSIBLE;
	}

	protected List<Object> constructQuery() {
		List<Object> query = new ArrayList<Object>(iclConstraint.getParameters().getList().size() * 2);
		for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
			Object minValue;
			Object maxValue;
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				minValue = getParameterValue(((ICLRange) parameter).getMin());
				maxValue = getParameterValue(((ICLRange) parameter).getMax());
			}
			else {
				minValue = getParameterValue((ICLValue) parameter);
				maxValue = minValue;
			}
			
			if(minValue == null || maxValue == null) {
				query.add(Index.UNDEFINED_PARAMETER);
				query.add(Index.UNDEFINED_PARAMETER);
			}
			else {
				query.add(minValue);
				query.add(maxValue);
			}
			
		}
		return query;
	}

	/**
	 * Check whether the matching action can be executed, i.e.
	 * all required dependencies have already been bound.
	 * 
	 * @return	True if the action is executable.
	 */
	protected boolean isExecutable() {
		//TODO if an unbound dependency appears both as a simple parameter and a range/attribute, this should return true
		for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
			/*
			 * Both minimum and maximum of each range need to be bound to execute,
			 * since the index can not be used to bind these.
			 */
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				if(getParameterValue(((ICLRange) parameter).getMin()) == null ||
						getParameterValue(((ICLRange) parameter).getMax()) == null) {
					return false;
				}
			}
			/*
			 * if there is an unbound attribute dependency,
			 * the index can also not be used to bind it
			 */
			else if(parameter.eClass() == ICLPackage.Literals.ICL_VARIABLE_ATTRIBUTE) {
				if(getParameterValue(((ICLValue) parameter)) == null) {
					return false;
				}
			}
		}
		return true;
	}

	public PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getPatternNode(ICLVariable parameter) {
		return iclValueToPatternNode.get(parameter).get(0);
	}

	@SuppressWarnings("unchecked")
	@Override
	public MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePristine(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> clonePatternConstraint,
			Map<Object, Object> clonePNs) {
		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNodeClones
				= new HashMap<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		for(Entry<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> entry:iclValueToPatternNode.entrySet()) {
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> patternNodeClones =
					new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:entry.getValue()) {
				patternNodeClones.add((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(pn));
			}
			iclValueToPatternNodeClones.put(entry.getKey(), patternNodeClones);
		}
		
		List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> requiredClones =
				new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> requirement:getRequiredBindings()) {
			requiredClones.add((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clonePNs.get(requirement));
		}
		
		return new MLSDMIndexStepMatchingAction(id, matcher, iclConstraint, expression, iclValueToPatternNodeClones, requiredClones);
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "match\t\t(" + expression.getExpressionString() + ")";
	}

}
