package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternConstraint;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternNode;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionCheckMatchingAction;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

/**
 * Default implementation of the search model builder for the MLSDM
 * meta model.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMEmulateDefaultSearchModelBuilder extends MLSDMSearchModelBuilder{

	/* (non-Javadoc)
	 * @see de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBuilder#createSPOExpressionPatternConstraints(de.mdelab.mlstorypatterns.AbstractStoryPatternObject, de.mdelab.mlexpressions.MLExpression, java.util.Map, java.util.Map)
	 */
	@Override
	protected Collection<MLSDMPatternConstraint> createSPOExpressionPatternConstraints(AbstractStoryPatternObject spo, MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes, Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		return Collections.singletonList(createGenericSPOExpressionPatternConstraint(spo, expression, patternNodes));
	}	

	/* (non-Javadoc)
	 * @see de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBuilder#createExpressionPatternConstraints(de.mdelab.mlexpressions.MLExpression, java.util.Map, java.util.Map, de.mdelab.sdm.interpreter.core.variables.VariablesScope)
	 */
	protected Collection<MLSDMPatternConstraint> createExpressionPatternConstraints(MLExpression expression, Map<String, AbstractStoryPatternObject> nonCreateSPOs, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope) {
		return Collections.singletonList(createGenericExpressionPatternConstraint(expression, nonCreateSPOs, patternNodes, dummyVariablesScope));
	}

	/* (non-Javadoc)
	 * @see de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMSearchModelBuilder#createGenericExpressionPatternConstraint(de.mdelab.mlexpressions.MLExpression, java.util.Map, java.util.Map, de.mdelab.sdm.interpreter.core.variables.VariablesScope)
	 */
	@Override
	protected MLSDMPatternConstraint createGenericExpressionPatternConstraint(MLExpression expression,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope) {

		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(expression);
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> dependencies
					= new HashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(AbstractStoryPatternObject spo:nonCreateSPOs.values()) {
			dependencies.add(patternNodes.get(spo));
		}
		
		patternConstraint.addDependencies(dependencies);
		
		MLSDMExpressionCheckMatchingAction matchingAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, dependencies);
		patternConstraint.addPossibleAction(matchingAction);
		patternConstraint.setExplicitCheckAction(matchingAction);
		
		return patternConstraint;
	}

}
