package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.MlsdmPackage;
import de.mdelab.mlsdm.StoryNode;
import de.mdelab.mlsdm.interpreter.expressions.MLSDMExpressionInterpreterManager;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMAdapterBackedOCLExpressionInterpreter;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternObject;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMMetadataIndex extends MLSDMReferenceIndex {

	public static final int EXPRESSION_NOT_REGISTERED = -1;
	private Map<EReference, Integer> maxCardinalities;
	private Map<EReference, Integer> maxInverseCardinalities;
	private Map<EReference, Integer> minCardinalities;
	private Map<EReference, Integer> minInverseCardinalities;
	private Map<MLExpression, Integer> validInstanceNumbers;		//TODO this needs to be maintained somehow
	private MLSDMExpressionInterpreterManager expressionInterpreterManager;
	
	public MLSDMMetadataIndex() {
		super();
		this.maxCardinalities = new HashMap<EReference, Integer>();
		this.maxInverseCardinalities = new HashMap<EReference, Integer>();
		this.minCardinalities = new HashMap<EReference, Integer>();
		this.minInverseCardinalities = new HashMap<EReference, Integer>();
		this.validInstanceNumbers = new HashMap<MLExpression, Integer>();
		this.expressionInterpreterManager = new MLSDMExpressionInterpreterManager(MLSDMMetadataIndex.class.getClassLoader());
		try {
			this.expressionInterpreterManager.registerExpressionInterpreter(new MLSDMAdapterBackedOCLExpressionInterpreter(this), "OCL", "1.0");
		} catch (SDMException e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public void addReference(EObject source, EReference reference, EObject target) {
		super.addReference(source, reference, target);
		
		int cardinality;
		if(reference.isMany()) {
			cardinality = ((Collection<?>) source.eGet(reference)).size();
		}
		else {
			cardinality = 1;
		}
		if(!maxCardinalities.containsKey(reference) || cardinality > maxCardinalities.get(reference)) {
			maxCardinalities.put(reference, cardinality);
		}
//		if(!minCardinalities.containsKey(reference) || cardinality < minCardinalities.get(reference)) {	//TODO obviously cannot be maintained this way...
//			minCardinalities.put(reference, cardinality);
//		}

		int inverseCardinality = getInverseReferences(target, reference).size();
		if(!maxInverseCardinalities.containsKey(reference) || inverseCardinality > maxInverseCardinalities.get(reference)) {
			maxInverseCardinalities.put(reference, inverseCardinality);
		}
//		if(!minInverseCardinalities.containsKey(reference) || inverseCardinality < minInverseCardinalities.get(reference)) {	//TODO obviously cannot be maintained this way...
//			minInverseCardinalities.put(reference, inverseCardinality);
//		}
	}
	
	public int getMaxCardinality(EReference reference) {
		return maxCardinalities.containsKey(reference) ? maxCardinalities.get(reference) : 0;
	}

	public int getMinCardinality(EReference reference) {
		return minCardinalities.containsKey(reference) ? minCardinalities.get(reference) : 0;
	}

	public int getMaxInverseCardinality(EReference reference) {
		return maxInverseCardinalities.containsKey(reference) ? maxInverseCardinalities.get(reference) : 0;
	}

	public int getMinInverseCardinality(EReference reference) {
		return minInverseCardinalities.containsKey(reference) ? minInverseCardinalities.get(reference) : 0;
	}
	
	public int getValidInstanceNumbers(Object expression) {
		if(!validInstanceNumbers.containsKey(expression)) {
			return EXPRESSION_NOT_REGISTERED;
		}
		else {
			return validInstanceNumbers.get(expression);
		}
	}
	
	public void registerExpression(MLExpression expression, EClass contextClass,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs) {
		Collection<EObject> domain = getDomain(contextClass);
		int validInstanceNumber = 0;
		for(EObject context:domain) {
			try {
				if((Boolean) expressionInterpreterManager.evaluateExpression(expression, contextClass, context, vs).getValue()) {
					validInstanceNumber++;
				}
			} catch (SDMException e) {
				e.printStackTrace();
			}
		}
		validInstanceNumbers.put(expression, validInstanceNumber);
	}

	public void registerExpressions(Activity activity, Collection<Variable<EClassifier>> inputVariables, boolean print) {
		VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs
				= new VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(
						new NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>());
		for(Variable<EClassifier> variable:inputVariables) {
			vs.createVariable(variable.getName(), variable.getClassifier(), variable.getValue());
		}
		
		for(ActivityNode node:activity.getNodes()) {
			if(node.eClass() == MlsdmPackage.Literals.STORY_NODE) {
				registerExpressions(((StoryNode)node).getStoryPattern(), vs, print);
			}
		}
	}

	public void registerExpressions(StoryPattern pattern,
			Collection<Variable<EClassifier>> variables,
			boolean print) {
		VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs
		= new VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(
				new NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>());
		for(Variable<EClassifier> variable:variables) {
			vs.createVariable(variable.getName(), variable.getClassifier(), variable.getValue());
		}
		
		registerExpressions(pattern, vs, print);
	}
	
	public void registerExpressions(StoryPattern pattern,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> vs,
			boolean print) {
		for(AbstractStoryPatternObject spo:pattern.getStoryPatternObjects()) {
			if(!(spo.getType().eClass() == EcorePackage.Literals.ECLASS)) {
				continue;
			}
			for(MLExpression expression:((StoryPatternObject) spo).getConstraints()) {
				long start = System.currentTimeMillis();
				registerExpression(expression, (EClass) spo.getType(), vs);
				long end = System.currentTimeMillis();
				if(print) System.out.println(spo.getName() + "_expression_time=" + (end - start));
			}
		}
	}

}
