package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlexpressions.MlexpressionsFactory;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.searchModel.MLSDMSearchModelBasedInterpreter;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternConstraint;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMPatternNode;
import de.mdelab.mlsdm.interpreter.searchModel.model.MLSDMSearchModel;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMContainerMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMContainmentCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMContainmentMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionLinkCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexArbitraryMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexLastParameterMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexStagedMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkCheckOptimizedMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkContainmentReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkTraverseReverseMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMMapCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMMapEntryMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMMapValueMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMNACCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMSPOExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions.MLSDMICLExpressionAnalyzer;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.ContainmentLink;
import de.mdelab.mlstorypatterns.ExpressionLink;
import de.mdelab.mlstorypatterns.MapEntryLink;
import de.mdelab.mlstorypatterns.MatchTypeEnum;
import de.mdelab.mlstorypatterns.MlstorypatternsPackage;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLExpression;
import de.mdelab.sdm.icl.iCL.ICLFactory;
import de.mdelab.sdm.icl.iCL.ICLPackage;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLParameterList;
import de.mdelab.sdm.icl.iCL.ICLRange;
import de.mdelab.sdm.icl.iCL.ICLStringExpression;
import de.mdelab.sdm.icl.iCL.ICLValue;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.icl.iCL.ICLVariableAttribute;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBasedMatcher;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBuilder;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

/**
 * Default implementation of the search model builder for the MLSDM
 * meta model.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMSearchModelBuilder extends SearchModelBuilder<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>{

	/*
	 * The matcher for which to create search models.
	 */
	protected MLSDMSearchModelBasedMatcher matcher;

	/*
	 * Identifier to assign to the next created matching action;
	 */
	protected int currentActionId;
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBuilder#setMatcher(de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBasedMatcher)
	 */
	@Override
	public void setMatcher(
			SearchModelBasedMatcher<?, ?, ?, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> matcher) {
		super.setMatcher(matcher);
		if(matcher instanceof MLSDMSearchModelBasedMatcher) {
			this.matcher = (MLSDMSearchModelBasedMatcher) matcher;	//TODO maybe this can be avoided
		}
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModelBuilder#createSearchModel(java.lang.Object)
	 */
	@Override
	public MLSDMSearchModel createSearchModel(StoryPattern storyPattern) {
		currentActionId = 0;
		
		MLSDMSearchModel searchModel = new MLSDMSearchModel();
		searchModel.setName(storyPattern.getName());
		
		Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes = createPatternNodes(searchModel, storyPattern);
		Collection<MLSDMPatternConstraint> patternConstraints = createPatternConstraints(storyPattern, patternNodes);
		int constraintId = 0;
		for(MLSDMPatternConstraint pc:patternConstraints) {
			pc.setId(constraintId);
			constraintId++;
		}
		
		List<MLSDMPatternNode> nodeList = new ArrayList<MLSDMPatternNode>(patternNodes.values());
		nodeList.sort(new Comparator<MLSDMPatternNode>() {

			@Override
			public int compare(MLSDMPatternNode o1, MLSDMPatternNode o2) {
				return Integer.compare(o1.getId(), o2.getId());
			}
			
		});
		searchModel.addPatternNodes(nodeList);
		searchModel.addPatternConstraints(patternConstraints);
		return searchModel;
	}
	
	/**
	 * Create the pattern nodes for a story pattern.
	 * 
	 * @param searchModel
	 * 			The search model in which to create the pattern nodes.
	 * @param storyPattern
	 * 			The story pattern.
	 * @return	A mapping from story pattern objects to the pattern node
	 * 			they are represented by.
	 */
	protected Map<AbstractStoryPatternObject, MLSDMPatternNode> createPatternNodes(MLSDMSearchModel searchModel, StoryPattern storyPattern) {
		Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes = new HashMap<AbstractStoryPatternObject, MLSDMPatternNode>();
		int id = 0;
		for(AbstractStoryPatternObject spo:matcher.getSpFacade().getStoryPatternObjects(storyPattern)) {
			if(matcher.getSpoFacade().isCreate(spo) && !matcher.getSpoFacade().isOptional(spo)) {
				continue;
			}
			patternNodes.put(spo, new MLSDMPatternNode(id++, searchModel, spo));
		}
		return patternNodes;
	}
	
	/**
	 * Create the pattern constraints for a story pattern.
	 * 
	 * @param storyPattern
	 * 			The story pattern.
	 * @param patternNodes
	 * 			A collection containing all created pattern constraints.
	 * @return
	 */
	protected Collection<MLSDMPatternConstraint> createPatternConstraints(StoryPattern storyPattern, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		patternConstraints.addAll(createLinkPatternConstraints(storyPattern, patternNodes));
		patternConstraints.addAll(createConstraintPatternConstraints(storyPattern, patternNodes));
		patternConstraints.addAll(createNACPatternConstraints(storyPattern, patternNodes));
		
		if(!matcher.getInterpreterParameters().containsKey(MLSDMSearchModelBasedInterpreter.ENABLE_DOMAIN_MATCHING) || (Boolean)matcher.getInterpreterParameters().get(MLSDMSearchModelBasedInterpreter.ENABLE_DOMAIN_MATCHING)) {
			patternConstraints.addAll(createDomainPatternConstraints(storyPattern, patternNodes));
		}
		
		return patternConstraints;
	}
	
	/**
	 * Create pattern constraints based on edges in the story pattern.
	 * 
	 * @param storyPattern
	 * 			The story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	A collection of pattern constraints representing edge
	 * 			constraints in the story pattern.
	 */
	protected Collection<MLSDMPatternConstraint> createLinkPatternConstraints(StoryPattern storyPattern, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();

		for(AbstractStoryPatternLink link:storyPattern.getStoryPatternLinks()) {
			/*
			 * Only consider links that are not part of a NAC
			 * and that are not being created.
			 */
			if((link.getModifier() != ModifierEnum.CREATE || link.getMatchType() == MatchTypeEnum.OPTIONAL) && 
					(link.getSource().getStoryPattern() == storyPattern || link.getSource().getStoryPattern() == storyPattern.getContainingPattern()) &&
					(link.getTarget().getStoryPattern() == storyPattern || link.getTarget().getStoryPattern() == storyPattern.getContainingPattern())) {
				final EClass linkEClass = link.eClass();
				
				MLSDMPatternConstraint patternConstraint = null;
				if (linkEClass == MlstorypatternsPackage.Literals.STORY_PATTERN_LINK) {
					patternConstraint = createLinkPatternConstraint((StoryPatternLink) link, patternNodes);
				}
				else if(linkEClass == MlstorypatternsPackage.Literals.CONTAINMENT_LINK) {
					patternConstraint = createContainmentLinkPatternConstraint((ContainmentLink) link, patternNodes);
				}
				else if(linkEClass == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK) {
					patternConstraint = createMapEntryLinkPatternConstraint((MapEntryLink) link, patternNodes);
				}
				else if(linkEClass == MlstorypatternsPackage.Literals.EXPRESSION_LINK) {
					patternConstraint = createExpressionLinkPatternConstraint((ExpressionLink) link, patternNodes);
				}
				
				if(patternConstraint != null){
					patternConstraint.setOptional(link.getMatchType() == MatchTypeEnum.OPTIONAL);
					patternConstraints.add(patternConstraint);
				}
			}
		}
		return patternConstraints;
	}
	
	/**
	 * Create a pattern constraint for a simple edge (link) in
	 * the story pattern.
	 * 
	 * @param link
	 * 			The edge for which to create a pattern constraint.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createLinkPatternConstraint(StoryPatternLink link,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(link);
		MLSDMPatternNode sourceNode = patternNodes.get(link.getSource());
		MLSDMPatternNode targetNode = patternNodes.get(link.getTarget());
		patternConstraint.addDependency(sourceNode);
		patternConstraint.addDependency(targetNode);
		
		/*
		 * Create action for explicit check; use optimized version if possible.
		 */
		MLSDMLinkCheckMatchingAction explicitCheckAction;
		if(matcher.getReferenceAdapter() != null) {
			explicitCheckAction = new MLSDMLinkCheckOptimizedMatchingAction(currentActionId++, matcher, patternConstraint, link, sourceNode, targetNode);
		}
		else {
			explicitCheckAction = new MLSDMLinkCheckMatchingAction(currentActionId++, matcher, patternConstraint, link, sourceNode, targetNode);
		}
		patternConstraint.addPossibleAction(explicitCheckAction);
		patternConstraint.setExplicitCheckAction(explicitCheckAction);
		
		/*
		 * Create actions for implicit checks (link traversal or lookup).
		 */
		patternConstraint.addPossibleAction(new MLSDMLinkTraverseMatchingAction(currentActionId++, matcher, patternConstraint, link, link.getFeature(), sourceNode, targetNode));
		if(link.getFeature() instanceof EReference) {
			/*
			 * If this has a reverse reference, use it for reverse traversal
			 * otherwise use the reference adapter if available.
			 */
			if(((EReference) link.getFeature()).getEOpposite() != null) {
				patternConstraint.addPossibleAction(new MLSDMLinkTraverseMatchingAction(currentActionId++, matcher, patternConstraint, link, ((EReference) link.getFeature()).getEOpposite(), targetNode, sourceNode));
			}
			else if(((EReference) link.getFeature()).isContainment()) {
				patternConstraint.addPossibleAction(new MLSDMLinkContainmentReverseMatchingAction(currentActionId++, matcher, patternConstraint, link, link.getFeature(), sourceNode, targetNode));
			}
			else if(matcher.getReferenceAdapter() != null) {
				patternConstraint.addPossibleAction(new MLSDMLinkTraverseReverseMatchingAction(currentActionId++, matcher, link, link.getFeature(), sourceNode, targetNode));
			}
			
			if(matcher.getReferenceAdapter() != null) {
				patternConstraint.addPossibleAction(new MLSDMLinkLookupMatchingAction(currentActionId++, matcher, link, sourceNode, targetNode));
			}
		}
		
		return patternConstraint;
	}

	/**
	 * Create a pattern constraint for a containment link in the
	 * story pattern.
	 * 
	 * @param link
	 * 			The containment link.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createContainmentLinkPatternConstraint(ContainmentLink link,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(link);
		MLSDMPatternNode containingNode = patternNodes.get(link.getSource());
		MLSDMPatternNode containedNode = patternNodes.get(link.getTarget());
		
		patternConstraint.addDependency(containingNode);
		patternConstraint.addDependency(containedNode);
		
		/*
		 * Create action for the explicit check.
		 */
		MLSDMContainmentCheckMatchingAction explicitCheckAction = new MLSDMContainmentCheckMatchingAction(currentActionId++, matcher, link, containingNode, containedNode);
		patternConstraint.addPossibleAction(explicitCheckAction);
		patternConstraint.setExplicitCheckAction(explicitCheckAction);
		
		/*
		 * Create actions for implicit checks (containment link traversal).
		 */
		patternConstraint.addPossibleAction(new MLSDMContainmentMatchingAction(currentActionId++, matcher, link, containingNode, containedNode));
		patternConstraint.addPossibleAction(new MLSDMContainerMatchingAction(currentActionId++, matcher, link, containingNode, containedNode));
		
		return patternConstraint;
	}

	/**
	 * Create a pattern constraint for a map entry link in the
	 * story pattern.
	 * 
	 * @param link
	 * 			The map entry link.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createMapEntryLinkPatternConstraint(MapEntryLink link,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(link);
		MLSDMPatternNode sourceNode = patternNodes.get(link.getSource());
		MLSDMPatternNode keyNode = patternNodes.get(link.getTarget());
		MLSDMPatternNode valueNode = link.getValueTarget() != null ? patternNodes.get(link.getValueTarget()) : null;
		patternConstraint.addDependency(sourceNode);
		patternConstraint.addDependency(keyNode);
		if(valueNode != null) {
			patternConstraint.addDependency(valueNode);
		}
		
		/*
		 * Create action for explicit check.
		 */
		MLSDMMapCheckMatchingAction explicitCheckAction = new MLSDMMapCheckMatchingAction(currentActionId++, matcher, link, sourceNode, valueNode, keyNode);
		patternConstraint.addPossibleAction(explicitCheckAction);
		patternConstraint.setExplicitCheckAction(explicitCheckAction);
		
		/*
		 * Create actions for implicit checks (link traversal).
		 */
		if(valueNode != null) {
			patternConstraint.addPossibleAction(new MLSDMMapValueMatchingAction(currentActionId++, matcher, link, sourceNode, valueNode, keyNode));
		}
		patternConstraint.addPossibleAction(new MLSDMMapEntryMatchingAction(currentActionId++, matcher, link, sourceNode, valueNode, keyNode));
		
		return patternConstraint;
	}

	/**
	 * Create a pattern constraint for an expression link in the
	 * story pattern.
	 * 
	 * @param link
	 * 			The expression link.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createExpressionLinkPatternConstraint(ExpressionLink link,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(link);
		MLSDMPatternNode sourceNode = patternNodes.get(link.getSource());
		MLSDMPatternNode targetNode = patternNodes.get(link.getTarget());
		patternConstraint.addDependency(sourceNode);
		patternConstraint.addDependency(targetNode);
		
		/*
		 * Create action for explicit check.
		 */
		MLSDMExpressionLinkCheckMatchingAction explicitCheckAction = new MLSDMExpressionLinkCheckMatchingAction(currentActionId++, matcher, link, sourceNode, targetNode);
		patternConstraint.addPossibleAction(explicitCheckAction);
		patternConstraint.setExplicitCheckAction(explicitCheckAction);
		
		/*
		 * Create actions for implicit checks (link traversal).
		 */
		patternConstraint.addPossibleAction(new MLSDMExpressionLinkTraverseMatchingAction(currentActionId++, matcher, link, sourceNode, targetNode));
		
		return patternConstraint;
	}

	/**
	 * Create pattern constraints for all expression constraints in the
	 * story pattern. This includes both constraints formulated over the
	 * entire story pattern and over single story pattern objects.
	 * 
	 * @param storyPattern
	 * 			The story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	A collection of the created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createConstraintPatternConstraints(StoryPattern storyPattern, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		Map<String, AbstractStoryPatternObject> nonCreateSPOs = new HashMap<String, AbstractStoryPatternObject>();
		VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope = createDummyVariablesScope(matcher.getSpFacade().getStoryPatternObjects(storyPattern));
		matcher.getExpressionAnalyzerManager().setVariablesScope(dummyVariablesScope);
		
		/*
		 * Create mappings between all story pattern objects without
		 * modifier CREATE and their name.
		 */
		for(AbstractStoryPatternObject spo:matcher.getSpFacade().getStoryPatternObjects(storyPattern)) {
			if(!matcher.getSpoFacade().isCreate(spo)) {
				nonCreateSPOs.put(spo.getName(), spo);
			}
		}
		
		/*
		 * Create pattern constraints for all expression constraints
		 * formulated over a single story pattern object.
		 */
		for(AbstractStoryPatternObject spo:matcher.getSpFacade().getStoryPatternObjects(storyPattern)) {
			if(!matcher.getSpoFacade().isCreate(spo)) {
				for(MLExpression expression:matcher.getSpoFacade().getConstraints(spo)) {
					Collection<MLSDMPatternConstraint> spoPatternConstraints = createSPOExpressionPatternConstraints(spo, expression, patternNodes, nonCreateSPOs);
					patternConstraints.addAll(spoPatternConstraints);
				}
			}
		}
		
		/*
		 * Create pattern constraints for all expression constraints
		 * formulated over the entire story pattern.
		 */
		for(MLExpression expression:matcher.getSpFacade().getConstraints(storyPattern)) {
			Collection<MLSDMPatternConstraint> expressionPatternConstraints = createExpressionPatternConstraints(expression, nonCreateSPOs, patternNodes, dummyVariablesScope);
			patternConstraints.addAll(expressionPatternConstraints);
		}
		
		return patternConstraints;
	}

	/**
	 * Create a dummy variables scope to use in the analysis of expression
	 * constraints. It has to include dummy variables for all story pattern
	 * objects in the story pattern.
	 * 
	 * @param storyPatternObjects
	 * 			A collection of story pattern objects to create dummy
	 * 			variables for.
	 * @return	The created dummy variables scope.
	 */
	protected VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> createDummyVariablesScope(
			Collection<AbstractStoryPatternObject> storyPatternObjects) {
		VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope
				= new VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(new NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>(), matcher.getVariablesScope(), Collections.<String, Map<String, List<String>>>emptyMap());
		for(AbstractStoryPatternObject spo:storyPatternObjects) {
			dummyVariablesScope.createVariable(spo.getName(), spo.getType(), null);
		}
		
		return dummyVariablesScope;
	}

	/**
	 * Create a pattern constraint based on an expression constraint
	 * formulated over a given story pattern object. Note that this
	 * allows the creation of multiple pattern constraints for a single
	 * expression constraint.
	 * 
	 * @param spo
	 * 			The story pattern object.
	 * @param expression
	 * 			The expression constraint.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createSPOExpressionPatternConstraints(AbstractStoryPatternObject spo, MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes, Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		/*
		 * Check if there is an expression analyzer that can parse the expression's AST
		 * and if there is special handling for the expression's language.
		 */
		String language = expression.getExpressionLanguage();
		if(matcher.getExpressionAnalyzerManager().getAnalyzer(language) != null) {
			Object ast = matcher.getFacadeFactory().getExpressionFacade().getExpressionAST(expression);
			if(ast == null) {
				ast = matcher.getExpressionAnalyzerManager().parseAST(expression, spo.getType());
				matcher.getFacadeFactory().getExpressionFacade().setExpressionAST(expression, ast);
			}
			if(language.equals("ICL")){
				return createSPOIndexPatternConstraints(spo, expression, (ICLExpression) ast, patternNodes, nonCreateSPOs);
			}
		}
		
		/*
		 * If no ast could be parsed or the expression language has no special handling,
		 * return a generic pattern constraint.
		 */
		return Collections.singletonList(createGenericSPOExpressionPatternConstraint(spo, expression, patternNodes));
	}

	/**
	 * Create pattern constraints based on an index constraint over a
	 * single story pattern object.
	 * 
	 * @param spo
	 * 			The story pattern object.
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param iclExpression
	 * 			The AST of the expression representing the index constraint.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createSPOIndexPatternConstraints(AbstractStoryPatternObject spo, MLExpression expression, ICLExpression iclExpression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes, Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		if(iclExpression.eClass() == ICLPackage.Literals.ICL_CONSTRAINT && ((ICLConstraint) iclExpression).getOperation().getName().equals("CONTAINS")) {
			return createIndexContainmentPatternConstraints((ICLConstraint) iclExpression, expression, patternNodes, nonCreateSPOs);
		}
		else {
			return Collections.singletonList(createGenericSPOExpressionPatternConstraint(spo, expression, patternNodes));
		}
	}
	

	/**
	 * Create pattern constraints based on an index constraint over the
	 * entire story pattern.
	 * 
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param iclExpression
	 * 			The AST of the expression representing the index constraint.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @param dummyVariablesScope
	 * 			A dummy variables scope with dummy variables for all story
	 * 			pattern objects in the story pattern.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createIndexPatternConstraints(MLExpression expression, ICLExpression iclExpression,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope) {
		if(iclExpression.eClass() == ICLPackage.Literals.ICL_CONSTRAINT && ((ICLConstraint) iclExpression).getOperation().getName().equals("CONTAINS")) {
			return createIndexContainmentPatternConstraints((ICLConstraint) iclExpression, expression, patternNodes, nonCreateSPOs);
		}
		else {
			return Collections.singletonList(createGenericExpressionPatternConstraint(expression, nonCreateSPOs, patternNodes, dummyVariablesScope));
		}
	}

	/**
	 * Create pattern constraints for an index containment constraint.
	 * 
	 * @param iclConstraint
	 * 			The AST of the expression representing the index constraint.
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createIndexContainmentPatternConstraints(ICLConstraint iclConstraint,
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes, Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		Index index = (Index) matcher.getVariablesScope().getVariable(iclConstraint.getIndexID()).getValue();
		switch(index.getAccessType()) {
		case ARBITRARY_KEY:
			return createArbitraryAccessIndexConstraints(iclConstraint, expression, patternNodes, nonCreateSPOs);
		case FULL_KEY:
			return createFullAccessIndexConstraints(iclConstraint, expression, patternNodes, nonCreateSPOs);
		case STAGED_KEY:
			return createStagedAccessIndexConstraints(iclConstraint, expression, patternNodes, nonCreateSPOs);
		default:
			return createFullAccessIndexConstraints(iclConstraint, expression, patternNodes, nonCreateSPOs);
		}
	}

	/**
	 * Create pattern constraints for an index constraint over an
	 * index with access type STAGED_KEY.
	 * 
	 * @param iclConstraint
	 * 			The AST of the expression representing the index constraint.
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createStagedAccessIndexConstraints(ICLConstraint iclConstraint,
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		
		if(matcher.getInterpreterParameters().containsKey(MLSDMSearchModelBasedInterpreter.ENABLE_STAGED_INDEX_MATCHING) && (Boolean) matcher.getInterpreterParameters().get(MLSDMSearchModelBasedInterpreter.ENABLE_STAGED_INDEX_MATCHING)) {	
			/*
			 * Create pattern constraints for staged matching based
			 * on the index.
			 */
			for(int i = 0; i < iclConstraint.getParameters().getList().size(); i++) {
				List<ICLParameter> derivedParameters = new ArrayList<ICLParameter>();
				for(int j = 0; j <= i; j++) {
					ICLParameter parameter = iclConstraint.getParameters().getList().get(j);
					derivedParameters.add(EcoreUtil.copy(parameter));
				}
				
				/*
				 * Derive constraints for virtual indices.
				 */
				ICLConstraint derivedConstraint = ICLFactory.eINSTANCE.createICLConstraint();
				derivedConstraint.setIndexID(iclConstraint.getIndexID());
				derivedConstraint.setOperation(EcoreUtil.copy(iclConstraint.getOperation()));
				
				ICLParameterList derivedParameterList = ICLFactory.eINSTANCE.createICLParameterList();
				derivedParameterList.getList().addAll(derivedParameters);
				derivedConstraint.setParameters(derivedParameterList);
				
				MLStringExpression derivedExpression = MlexpressionsFactory.eINSTANCE.createMLStringExpression();
				derivedExpression.setExpressionLanguageID(expression.getExpressionLanguage());
				MLSDMICLExpressionAnalyzer analyzer = (MLSDMICLExpressionAnalyzer) matcher.getExpressionAnalyzerManager().getAnalyzer(expression.getExpressionLanguage());
				derivedExpression.setExpressionString(analyzer.getExpressionString(derivedConstraint));
				
				patternConstraints.add(createStagedAccessIndexConstraint(derivedConstraint, derivedExpression, patternNodes, nonCreateSPOs));
			}
		}
		else {
			/*
			 * Create pattern constraints for regular index-based
			 * matching.
			 */
			MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(iclConstraint);

			Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
						mapICLValuesToPatternNodes(iclConstraint, nonCreateSPOs, patternNodes);
			
			patternConstraint.addDependencies(flatten(iclValueToPatternNode.values()));
			
			MLSDMIndexStagedMatchingAction matchingAction = new MLSDMIndexStagedMatchingAction(currentActionId++, matcher, iclConstraint, (MLStringExpression) expression, iclValueToPatternNode);
			patternConstraint.addPossibleAction(matchingAction);
			
			MLSDMExpressionCheckMatchingAction explicitCheckAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternConstraint.getDependencies());
			patternConstraint.setExplicitCheckAction(explicitCheckAction);
			
			patternConstraints.add(patternConstraint);
		}
		return patternConstraints;
	}

	protected MLSDMPatternConstraint createStagedAccessIndexConstraint(ICLConstraint iclConstraint,		//TODO maybe this can be removed.
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(iclConstraint);
		
		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
					mapICLValuesToPatternNodes(iclConstraint, nonCreateSPOs, patternNodes);
		
		patternConstraint.addDependencies(flatten(iclValueToPatternNode.values()));
		
		int parameterNumber = iclConstraint.getParameters().getList().size();
		if(iclConstraint.getParameters().getList().get(parameterNumber - 1).eClass() == ICLPackage.Literals.ICL_VARIABLE) {
			MLSDMIndexLastParameterMatchingAction matchingAction = new MLSDMIndexLastParameterMatchingAction(currentActionId++, matcher, iclConstraint, (MLStringExpression) expression, iclValueToPatternNode);
			patternConstraint.addPossibleAction(matchingAction);
		}
		
		MLSDMExpressionCheckMatchingAction expicitCheckAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternConstraint.getDependencies());
		patternConstraint.setExplicitCheckAction(expicitCheckAction);
		patternConstraint.addPossibleAction(expicitCheckAction);
		return patternConstraint;
	}

	/**
	 * Create pattern constraints based on an index constraint over an
	 * index with access type FULL_KEY.
	  * 
	 * @param iclConstraint
	 * 			The AST of the expression representing the index constraint.
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createFullAccessIndexConstraints(ICLConstraint iclConstraint,
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(iclConstraint);

		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
					mapICLValuesToPatternNodes(iclConstraint, nonCreateSPOs, patternNodes);
		
		patternConstraint.addDependencies(flatten(iclValueToPatternNode.values()));
		
		//TODO this should actually also allow matching all indexed nodes at once if applicable
		
		MLSDMExpressionCheckMatchingAction matchingAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternConstraint.getDependencies());
		patternConstraint.addPossibleAction(matchingAction);
		patternConstraint.setExplicitCheckAction(matchingAction);
		
		return Collections.singleton(patternConstraint);
	}

	/**
	 * Create pattern constraints based on an index constraint over an
	 * index with access type ARBITRARY_KEY.
	  * 
	 * @param iclConstraint
	 * 			The AST of the expression representing the index constraint.
	 * @param expression
	 * 			The representation of the expression formulating the index
	 * 			constraint in the story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createArbitraryAccessIndexConstraints(ICLConstraint iclConstraint,
			MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(iclConstraint);

		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
					mapICLValuesToPatternNodes(iclConstraint, nonCreateSPOs, patternNodes);
		
		patternConstraint.addDependencies(flatten(iclValueToPatternNode.values()));
		
		MLSDMIndexArbitraryMatchingAction matchingAction = new MLSDMIndexArbitraryMatchingAction(currentActionId++, matcher, iclConstraint, (MLStringExpression) expression, iclValueToPatternNode);
		patternConstraint.addPossibleAction(matchingAction);
	
		MLSDMExpressionCheckMatchingAction explicitCheckAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternConstraint.getDependencies());
		patternConstraint.setExplicitCheckAction(explicitCheckAction);
		
		return Collections.singleton(patternConstraint);
	}

	/**
	 * Create a mapping for values in parameters of an index constraints
	 * to pattern nodes in a search model. Note that an ICL value may be
	 * mapped to multiple pattern nodes, e.g. if the ICL value is given
	 * by an expression in another expression language.
	 * 
	 * @param iclConstraint
	 * 			The AST of the expression representing the index constraint.
	 * @param nonCreateSPOs
	 * 			A Collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return The created mapping.
	 */
	protected Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> mapICLValuesToPatternNodes(ICLConstraint iclConstraint, Map<String, AbstractStoryPatternObject> nonCreateSPOs, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		/*
		 * Add dependencies to the according pattern node for all parameters
		 * that refer to SPOs in this pattern.
		 */
		Map<ICLValue, List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> iclValueToPatternNode =
				new HashMap<ICLValue,  List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		for(ICLParameter parameter:iclConstraint.getParameters().getList()) {
			if(parameter.eClass() == ICLPackage.Literals.ICL_RANGE) {
				ICLRange rangeParameter = (ICLRange) parameter;
				
				/*
				 * Create mappings for the lower bound value if necessary.
				 */
				Set<MLSDMPatternNode> minParameterPatternNodes = new HashSet<MLSDMPatternNode>();
				Collection<String> minNames = getVariableNames(rangeParameter.getMin());
				for(String minName:minNames) {
					AbstractStoryPatternObject minSPO = nonCreateSPOs.get(minName);
					
					if(minSPO != null) {
						MLSDMPatternNode patternNode = patternNodes.get(minSPO);
						minParameterPatternNodes.add(patternNode);
					}
				}
				iclValueToPatternNode.put(rangeParameter.getMin(), new ArrayList<MLSDMPatternNode>(minParameterPatternNodes));
				
				/*
				 * Create mappings for the upper bound value if necessary.
				 */
				Set<MLSDMPatternNode> maxParameterPatternNodes = new HashSet<MLSDMPatternNode>();
				Collection<String> maxNames = getVariableNames(rangeParameter.getMax());
				for(String maxName:maxNames) {
					AbstractStoryPatternObject maxSPO = nonCreateSPOs.get(maxName);
					
					if(maxSPO != null) {
						MLSDMPatternNode patternNode = patternNodes.get(maxSPO);
						maxParameterPatternNodes.add(patternNode);
					}
				}
				iclValueToPatternNode.put(rangeParameter.getMax(), new ArrayList<MLSDMPatternNode>(maxParameterPatternNodes));
			}
			else {
				/*
				 * Create mappings for a single value parameter if necessary.
				 */
				Set<MLSDMPatternNode> parameterPatternNodes = new HashSet<MLSDMPatternNode>();
				Collection<String> names = getVariableNames((ICLValue) parameter);
				for(String name:names) {
					AbstractStoryPatternObject spo = nonCreateSPOs.get(name);
					
					if(spo != null) {
						MLSDMPatternNode patternNode = patternNodes.get(spo);
						parameterPatternNodes.add(patternNode);
					}
				}
				iclValueToPatternNode.put((ICLValue) parameter, new ArrayList<MLSDMPatternNode>(parameterPatternNodes));
			}
		}
		return iclValueToPatternNode;
	}
	
	/**
	 * Flatten a collection of collections of pattern nodes.
	 * 
	 * @param values
	 * 			The collection of collections of pattern nodes.
	 * @return	A set containing all pattern nodes contained in at least
	 * 			one of the input collections.
	 */
	protected Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> flatten(
			Collection<List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> values) {
		Set<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> dependencies =
				new HashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(List<? extends PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> value:values) {
			dependencies.addAll(value);
		}
		return dependencies;
	}

	/**
	 * Retrieve the names of all variables referenced in an ICL value.
	 * 
	 * @param value
	 * 			The ICL value.
	 * @return	A collection containing the names of all referenced
	 * 			variables.
	 */
	protected Collection<String> getVariableNames(ICLValue value) {
		if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE_ATTRIBUTE) {
			return Collections.singleton(((ICLVariableAttribute) value).getVariable().getName());
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_VARIABLE) {
			return Collections.singleton(((ICLVariable) value).getName());
		}
		else if(value.eClass() == ICLPackage.Literals.ICL_STRING_EXPRESSION) {
			return matcher.getExpressionAnalyzerManager().getVariableNames(((ICLStringExpression) value).getExpression());
		}
		else{
			return Collections.emptyList();
		}
	}

	/**
	 * Create a pattern constraint based on an expression constraint
	 * formulated over the entire story pattern. Note that this
	 * allows the creation of multiple pattern constraints for a single
	 * expression constraint.
	 * 
	 * @param expression
	 * 			The expression constraint.
	 * @param nonCreateSPOs
	 * 			A collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param dummyVariablesScope
	 * 			A dummy variables scope with dummy variables for all story
	 * 			pattern objects in the story pattern.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createExpressionPatternConstraints(MLExpression expression, Map<String, AbstractStoryPatternObject> nonCreateSPOs, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope) {
		/*
		 * Check if there is an expression analyzer that can parse the expression's AST
		 * and if there is special handling for the expression's language.
		 */
		String language = expression.getExpressionLanguage();
		if(matcher.getExpressionAnalyzerManager().getAnalyzer(language) != null) {
			Object ast = matcher.getFacadeFactory().getExpressionFacade().getExpressionAST(expression);
			if(ast == null) {
				ast = matcher.getExpressionAnalyzerManager().parseAST(expression, null);
				matcher.getFacadeFactory().getExpressionFacade().setExpressionAST(expression, ast);
			}	
			if(language.equals("ICL")){
				return createIndexPatternConstraints(expression, (ICLExpression) ast, patternNodes, nonCreateSPOs, dummyVariablesScope);
			}
		}
		
		/*
		 * If no AST could be parsed or the expression language has no special handling,
		 * return a generic pattern constraint.
		 */
		return Collections.singletonList(createGenericExpressionPatternConstraint(expression, nonCreateSPOs, patternNodes, dummyVariablesScope));
	}

	/**
	 * Create a generic pattern constraint based on an expression
	 * constraint formulated over a single story pattern object.
	 * The created pattern constraint only has a single matching
	 * action which performs an explicit check of the constraint.
	 * 
	 * @param spo
	 * 			The story pattern object.
	 * @param expression
	 * 			The AST of the expression representing the index constraint.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createGenericSPOExpressionPatternConstraint(AbstractStoryPatternObject spo, MLExpression expression, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(expression);
		PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternNode = patternNodes.get(spo);
		patternConstraint.addDependency(patternNode);
		
		MLSDMSPOExpressionCheckMatchingAction matchingAction = new MLSDMSPOExpressionCheckMatchingAction(currentActionId++, matcher, expression, patternNode);
		patternConstraint.addPossibleAction(matchingAction);
		patternConstraint.setExplicitCheckAction(matchingAction);
		
		return patternConstraint;
	}
	
	/**
	 * Create a generic pattern constraint based on an expression
	 * constraint formulated over the entire story pattern.
	 * The created pattern constraint only has a single matching
	 * action which performs an explicit check of the constraint.
	 * 
	 * @param expression
	 * 			The AST of the expression representing the index constraint.
	 * @param nonCreateSPOs
	 * 			A collection of all story pattern objects without the
	 * 			CREATE modifier.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @param dummyVariablesScope
	 * 			A dummy variables scope with dummy variables for all story
	 * 			pattern objects in the story pattern.
	 * @return	The created pattern constraint.
	 */
	protected MLSDMPatternConstraint createGenericExpressionPatternConstraint(MLExpression expression,
			Map<String, AbstractStoryPatternObject> nonCreateSPOs,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes,
			VariablesScope<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dummyVariablesScope) {

		MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(expression);
		
		/*
		 * Compute pattern node dependencies of the expression.
		 */
		Collection<String> dependencyNames = matcher.getExpressionAnalyzerManager().getVariableNames(expression);
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> dependencies
				= new HashSet<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		if(dependencyNames != null) {
			/*
			 * Add dependencies as retrieved by analyzer.
			 */
			for(String dependencyName:dependencyNames) {
				AbstractStoryPatternObject spo = nonCreateSPOs.get(dependencyName);
				if(spo != null) {
					dependencies.add(patternNodes.get(spo));
				}
			}
		}
		else {
			/*
			 * dependencies could not be analyzed; add all potential dependencies.
			 */
			for(AbstractStoryPatternObject spo:nonCreateSPOs.values()) {
				dependencies.add(patternNodes.get(spo));
			}
		}
		
		patternConstraint.addDependencies(dependencies);
		
		MLSDMExpressionCheckMatchingAction matchingAction = new MLSDMExpressionCheckMatchingAction(currentActionId++, matcher, expression, dependencies);
		patternConstraint.addPossibleAction(matchingAction);
		patternConstraint.setExplicitCheckAction(matchingAction);
		
		return patternConstraint;
	}
	
	/**
	 * Create pattern constraints based on negative application conditions
	 * in the story pattern.
	 * 
	 * @param storyPattern
	 * 			The story pattern.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createNACPatternConstraints(StoryPattern storyPattern,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		for(StoryPattern nacPattern:storyPattern.getNegativeApplicationConditions()) {
			MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(nacPattern);
			List<MLSDMPatternNode> dependencies = getDependencies(storyPattern, nacPattern, patternNodes);
			patternConstraint.addDependencies(dependencies);
			
			MLSDMNACCheckMatchingAction checkAction = new MLSDMNACCheckMatchingAction(currentActionId++, matcher, nacPattern, dependencies);
			patternConstraint.addPossibleAction(checkAction);
			patternConstraint.setExplicitCheckAction(checkAction);
			
			patternConstraints.add(patternConstraint);
		}
		return patternConstraints;
	}
	
	/**
	 * Retrieve dependencies of a negative application condition to
	 * story pattern objects in the parent story pattern. Dependencies
	 * are story pattern objects in the parent pattern which have an
	 * edge to or from a story pattern object in the negative pattern.
	 * Such objects are required to be bound in order to be able to
	 * check the negative application condition.
	 * 
	 * @param storyPattern
	 * 			The story pattern.
	 * @param nacPattern
	 * 			The negative application condition.
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	A list containing all dependencies of the negative
	 * 			application condition in the parent pattern, represented
	 * 			by their corresponding pattern nodes.
	 */
	private List<MLSDMPatternNode> getDependencies(StoryPattern storyPattern, StoryPattern nacPattern,
			Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		Set<MLSDMPatternNode> dependencies = new HashSet<MLSDMPatternNode>();		
		
		/*
		 * collect all relevant links, i.e. links within the nac pattern
		 * and links connected to nac pattern nodes
		 */
		Set<AbstractStoryPatternLink> consideredLinks = new HashSet<AbstractStoryPatternLink>();
		consideredLinks.addAll(nacPattern.getStoryPatternLinks());
		for(AbstractStoryPatternLink link:storyPattern.getStoryPatternLinks()) {
			if(link.getSource().getStoryPattern() == nacPattern || link.getTarget().getStoryPattern() == nacPattern) {
				consideredLinks.add(link);
			}
		}
		
		/*
		 * add dependencies for links within the pattern
		 */
		for(AbstractStoryPatternLink link:consideredLinks) {			
			if(link.getSource().getStoryPattern() == storyPattern) {
				dependencies.add(patternNodes.get(link.getSource()));
			}
			if(link.getTarget().getStoryPattern() == storyPattern) {
				dependencies.add(patternNodes.get(link.getTarget()));
			}
			
			if(link.eClass() == MlstorypatternsPackage.Literals.MAP_ENTRY_LINK) {
				MapEntryLink mapLink = (MapEntryLink) link;
				if(mapLink.getValueTarget().getStoryPattern() == storyPattern) {
					dependencies.add(patternNodes.get(mapLink.getValueTarget()));
				}
			}
		}
		
		return new ArrayList<MLSDMPatternNode>(dependencies);
	}
	
	/**
	 * Create pattern constraints based on domain constraints over nodes
	 * in the story pattern.
	 * 
	 * @param patternNodes
	 * 			A mapping from story pattern objects to the pattern node
	 * 			which they are represented by.
	 * @return	The created pattern constraints.
	 */
	protected Collection<MLSDMPatternConstraint> createDomainPatternConstraints(
			StoryPattern storyPattern, Map<AbstractStoryPatternObject, MLSDMPatternNode> patternNodes) {
		List<MLSDMPatternConstraint> patternConstraints = new ArrayList<MLSDMPatternConstraint>();
		
		/*
		 * Matching by domain is only possible if
		 * there is a reference adapter.
		 */
		if(matcher.getReferenceAdapter() == null) {
			return patternConstraints;
		}
		
		for(AbstractStoryPatternObject spo:storyPattern.getStoryPatternObjects()) {
			if(!(spo.getType() instanceof EClass) ||
					(spo.getModifier() == ModifierEnum.CREATE &&
					!(spo.getMatchType() == MatchTypeEnum.OPTIONAL))) {
				continue;
			}
			MLSDMPatternNode patternNode = patternNodes.get(spo);
			
			MLSDMPatternConstraint patternConstraint = new MLSDMPatternConstraint(patternNode.getSpo().getType());
			patternConstraint.addDependency(patternNode);
			
			MLSDMDomainCheckMatchingAction checkAction = new MLSDMDomainCheckMatchingAction(currentActionId++, matcher, patternNode);
			patternConstraint.setExplicitCheckAction(checkAction);
			
			patternConstraint.addPossibleAction(checkAction);
			patternConstraint.addPossibleAction(new MLSDMDomainMatchMatchingAction(currentActionId++, matcher, patternNode));
			

			patternConstraint.setOptional(patternNode.getSpo().getMatchType() == MatchTypeEnum.OPTIONAL);
			
			patternConstraints.add(patternConstraint);
		}
		return patternConstraints;
	}

	public void setMatcher(MLSDMSearchModelBasedMatcher matcher) {
		this.matcher = matcher;
	}

}
