package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions;

import java.util.Collection;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.ocl.ParserException;
import org.eclipse.ocl.ecore.OCL;
import org.eclipse.ocl.ecore.OCLExpression;
import org.eclipse.ocl.ecore.OCL.Helper;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.MLStringExpression;
import de.mdelab.mlexpressions.MlexpressionsPackage;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzerManager;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.FeatureAccess;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;
import de.mdelab.sdm.interpreter.ocl.environment.CustomOCLEnvironmentFactory;

/**
 * This is an expression analyzer implementation for the
 * Object Constraint Language (OCL).
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMOCLExpressionAnalyzer extends MLSDMExpressionAnalyzer {

	/*
	 * OCL environment factory.
	 */
	private CustomOCLEnvironmentFactory environmentFactory;
	
	/*
	 * OCL object.
	 */
	private OCL ocl;
	
	/*
	 * Helper for parsing OCL expressions.
	 */
	private Helper helper;
	
	public MLSDMOCLExpressionAnalyzer(){
		this(null);	//TODO factory
	}
	
	public MLSDMOCLExpressionAnalyzer(ExpressionAnalyzerManager<EClassifier, EStructuralFeature, MLExpression> analyzerManager) {
		super(analyzerManager);
		ocl = OCL.newInstance();
		helper = ocl.createOCLHelper();
		helper.setContext(EcorePackage.Literals.EOBJECT);
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#getVariableNames(java.lang.Object)
	 */
	@Override
	public Collection<String> getVariableNames(MLExpression expression) {
		if(!(expression.eClass() == MlexpressionsPackage.Literals.ML_STRING_EXPRESSION) || !expression.getExpressionLanguage().equals("OCL")) {
			return null;
		}
		
		String expressionString = ((MLStringExpression) expression).getExpressionString();
		try {
			OCLExpression oclExpression = helper.createQuery(expressionString);
			
			MLSDMOCLVariableCollectorVisitor visitor = new MLSDMOCLVariableCollectorVisitor();
			oclExpression.accept(visitor);
			return visitor.getVariableNames();
		} catch (ParserException e) {
			e.printStackTrace();
		}		
		return null;
	}
	
	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#setVariablesScope(de.mdelab.sdm.interpreter.core.variables.VariablesScope)
	 */
	public void setVariablesScope(
			VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope) {
		super.setVariablesScope(variablesScope);
		/*
		 * Set up the ocl-related helpers in the context of
		 * the new variables scope.
		 */
		this.environmentFactory = new CustomOCLEnvironmentFactory(variablesScope);
		this.ocl = OCL.newInstance(environmentFactory);
		this.helper = ocl.createOCLHelper();
		this.helper.setContext(EcorePackage.Literals.EOBJECT);
	}

	@Override
	public Collection<FeatureAccess<EStructuralFeature>> getFeatureAccesses(
			MLExpression expression, EClassifier contextClassifier) {
		if(!(expression.eClass() == MlexpressionsPackage.Literals.ML_STRING_EXPRESSION) || !expression.getExpressionLanguage().equals("OCL")) {
			return null;
		}
		
		String expressionString = ((MLStringExpression) expression).getExpressionString();
		try {
			helper.setContext(contextClassifier);
			OCLExpression oclExpression = helper.createQuery(expressionString);
			
			MLSDMOCLFeatureAccessCollectorVisitor visitor = new MLSDMOCLFeatureAccessCollectorVisitor();
			oclExpression.accept(visitor);
			return visitor.getFeatureAccesses();
		} catch (ParserException e) {
			e.printStackTrace();
		}		
		return null;
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.expressions.ExpressionAnalyzer#parseAST(java.lang.Object)
	 */
	@Override
	public Object parseAST(MLExpression expression, EClassifier contextClassifier) {
		if (contextClassifier == null)
		{
			contextClassifier = EcorePackage.Literals.EOBJECT;
		}

		final Helper helper = ocl.createOCLHelper();

		helper.setContext(contextClassifier);

		try {
			return helper.createQuery(((MLStringExpression) expression).getExpressionString());
		} catch (ParserException e) {
			e.printStackTrace();
		}
		return null;
	}

}
