package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.expressions;
import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.ocl.ecore.utilities.AbstractVisitor;
import org.eclipse.ocl.expressions.VariableExp;

/**
 * This is an implementation of the visitor in the visitor pattern
 * which can visit the abstract syntax tree of an OCL expression
 * and collects the names of all variables referenced in the expression
 * in the course of the visiting process.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMOCLVariableCollectorVisitor extends AbstractVisitor<Object> {

	/*
	 * Collection of all collected variable names.
	 */
	private Collection<String> variableNames;
	
	public MLSDMOCLVariableCollectorVisitor() {
		super();
		variableNames = new ArrayList<String>();
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.ocl.utilities.AbstractVisitor#visitVariableExp(org.eclipse.ocl.expressions.VariableExp)
	 */
	@Override
	public Object visitVariableExp(VariableExp<EClassifier, EParameter> vd) {
		variableNames.add(vd.getReferredVariable().getName());
		return super.visitVariableExp(vd);
	}
	
	public Collection<String> getVariableNames(){
		return variableNames;
	}

}
