package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy;

/**
 * This is a pattern constraint selection strategy that selects
 * the next pattern constraint to check in a pseudo-random fashion.
 * Upon creation, it can be parametrized with a seed to allow the
 * reproduction of the computed matching order for a given model.
 * 
 * @author Matthias.Barkowsky
 *
 */
public class MLSDMIndeterministicSelectionStrategy extends OrderProducingSelectionStrategy<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> {
	
	/*
	 * The random object to use for all pseudo-random computations.
	 */
	private Random random;
	
	public MLSDMIndeterministicSelectionStrategy() {
		this.random = new Random();
	}
	
	public MLSDMIndeterministicSelectionStrategy(long seed) {
		this.random = new Random(seed);
	}

	/* (non-Javadoc)
	 * @see de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.strategy.OrderProducingSelectionStrategy#selectNextConstraint()
	 */
	@Override
	protected PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> selectNextConstraint() {
		List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> executableNodes = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint:searchModel.getPatternConstraints()) {
			if(patternConstraint.isActive() || patternConstraint.getCardinality() == PatternConstraint.MATCHING_NOT_POSSIBLE) {
				continue;
			}
			else {
				executableNodes.add(patternConstraint);
			}
		}
		
		if(executableNodes.size() < 1) {
			return null;
		}
		else{
			int i = random.nextInt(executableNodes.size());
			return executableNodes.get(i);
		}
	}

}
