package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.grgen;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;

import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMReferenceIndex;

public class GRGENGeometricMeansAccumulator {

	protected MLSDMReferenceIndex referenceAdapter;
	protected Map<EReference, Double> forwardCosts;
	protected Map<EReference, Double> backwardCosts;
	
	public GRGENGeometricMeansAccumulator(MLSDMReferenceIndex referenceAdapter) {
		this.referenceAdapter = referenceAdapter;
		this.forwardCosts = new HashMap<EReference, Double>();
		this.backwardCosts = new HashMap<EReference, Double>();
	}
	
	public double getForwardCost(EReference reference) {
		if(!forwardCosts.containsKey(reference)) {
			forwardCosts.put(reference, computeForwardCost(reference));
		}
		return forwardCosts.get(reference);
	}
	
	protected double computeForwardCost(EReference reference) {
		if(!reference.isMany()) {
			return 1;
		}
		
		Collection<EObject> sources = referenceAdapter.getDomain(reference.getEContainingClass());
		if(sources.size() == 0) {
			return 1;
		}

		long count = 0;
		double productLog = 0;
		for(EObject source:sources) {
			@SuppressWarnings("unchecked")
			int outEdges = ((Collection<? extends EObject>) source.eGet(reference)).size();
			if(outEdges <= 0) {
				continue;
			}
			productLog += outEdges != 0 ? Math.log(outEdges) : 0;
			count++;
		}
		return Math.pow(Math.E, productLog / count);
	}

	public double getBackwardCost(EReference reference) {
		if(!backwardCosts.containsKey(reference)) {
			backwardCosts.put(reference, computeBackwardCost(reference));
		}
		return backwardCosts.get(reference);
	}
	
	protected double computeBackwardCost(EReference reference) {
		Collection<EObject> targets = referenceAdapter.getDomain(reference.getEReferenceType());
		if(targets.size() == 0) {
			return 1;
		}

		double productLog = 0;
		long count = 0;
		for(EObject target:targets) {
			int inEdges =  referenceAdapter.getInverseReferences(target, reference).size();
			if(inEdges <= 0) {
				continue;
			}
			productLog += inEdges != 0 ? Math.log(inEdges) : 0;
			count++;
		}
		return Math.pow(Math.E, productLog / count);
	}
	
}
