package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMCardinalityMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMCompactingCostStrategy
		extends
			MLSDMHybridCostStrategy {

	private static final double DYNAMIC_OFFSET = Math.pow(2, 128);
	
	protected MLSDMMatchingCostStrategy dynamicCostStrategy;
	
	public MLSDMCompactingCostStrategy(
			MLSDMHybridSelectionStrategy selectionStrategy,
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			CostModel costModel,
			VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope,
			MLSDMMetadataIndex metadataAdapter) {
		super(selectionStrategy, searchModel, costModel, variablesScope, metadataAdapter);
		this.dynamicCostStrategy = createMatchingCostStrategy();
	}

	protected MLSDMMatchingCostStrategy createMatchingCostStrategy() {
		return new MLSDMCardinalityMatchingCostStrategy();
	}

	@Override
	public double calculateMatchingCost(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint) {
		if(patternConstraint.flagIsSet(MLSDMCompactingSelectionStrategy.PC_COMPACTED)) {
			return dynamicCostStrategy.calculateMatchingCost(patternConstraint) * DYNAMIC_OFFSET;
		}
		else {
			return super.calculateMatchingCost(patternConstraint);
		}
	}
}
