package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMDomainMatchMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMLinkLookupMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMSPOExpressionCheckMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.CheckMatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMCompactingSelectionStrategy
		extends
			MLSDMHybridSelectionStrategy {
	
	public static final int PC_COMPACTED = 0;
	
	private Map<ICLConstraint, Boolean> indexFilters;
	private int[] idMappings;
	private SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> compactSearchModel;

	public MLSDMCompactingSelectionStrategy(
			MLSDMMetadataIndex metadataAdapter, CostModel costModel,
			VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super(metadataAdapter, costModel, variablesScope);
		this.indexFilters = new HashMap<ICLConstraint, Boolean>(); 
	}

	@Override
	protected MLSDMMatchingCostStrategy createMatchingCostStrategy() {
		return new MLSDMCompactingCostStrategy(this, compactSearchModel, costModel, variablesScope, metadataAdapter);
	}
	
	@Override
	protected void doInitialize() {
		compactSearchModel = compactSearchModel(searchModel);
		super.doInitialize();
	}
	
	@SuppressWarnings("unchecked")
	private SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> compactSearchModel(SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel) {
		Map<Object, Object> clones = searchModel.clonePristine();
		SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> compactSearchModel
				= (SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clones.get(searchModel);
		doCompactSearchModel(compactSearchModel);
		idMappings = new int[searchModel.getPatternNodes().size()];
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:searchModel.getPatternNodes()) {
			idMappings[pn.getId()] = ((PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clones.get(pn)).getId();
		}
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:searchModel.getPatternConstraints()) {
			if(((PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) clones.get(pc)).flagIsSet(PC_COMPACTED)) {
				pc.setFlag(PC_COMPACTED);
			}
		}
		return compactSearchModel;
	}

	private void doCompactSearchModel(
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> compactSearchModel) {
		Map<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>, List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> anchors =
				new HashMap<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>, List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
		List<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> mergeableConstraints =
				new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:compactSearchModel.getPatternConstraints()) {
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> anchorList = computeAnchors(pc);
			anchors.put(pc, anchorList);
			if(isMergeable(pc, anchorList)) {
				mergeableConstraints.add(pc);
			}
		}
		
		while(!mergeableConstraints.isEmpty()) {
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc = mergeableConstraints.remove(mergeableConstraints.size() - 1);
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> anchorList = anchors.get(pc);
			for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:pc.getDependencies()) {
				if(!anchorList.contains(pn)) {
					compactSearchModel.getPatternNodes().remove(pn);
					pn.setId(-1);
					
					for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dependant:pn.getDependantPatternConstraints()){
						compactSearchModel.getPatternConstraints().remove(dependant);
						dependant.setFlag(PC_COMPACTED);
					}
				}
			}
			
			if(!anchorList.isEmpty()) {
				PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> anchor = anchors.get(pc).get(0);
				anchor.getDependantPatternConstraints().remove(pc);

				//update dependants of anchor
				for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> candidate:anchor.getDependantPatternConstraints()) {
					
					List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> candidateAnchors = computeAnchors(candidate);
					boolean wasMergeable = anchors.get(candidate).size() < 2;
					anchors.put(candidate, candidateAnchors);
					if(!wasMergeable && isMergeable(candidate, candidateAnchors)) {
						mergeableConstraints.add(candidate);
					}
				}
			}
		}
		
		for(int i = 0; i < compactSearchModel.getPatternNodes().size(); i++) {
			compactSearchModel.getPatternNodes().get(i).setId(i);
		}
	}

	private boolean isMergeable(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc,
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> anchors) {
				return anchors.size() < 2 &&
						anchors.size() < pc.getDependencies().size() &&
						!filtersAnchor(pc, anchors) &&
						applicableFromAnchor(pc, anchors);
	}
	
	private boolean applicableFromAnchor(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> candidate,
			List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> candidateAnchors) {
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:candidate.getPossibleActions()) {
			if(ma.getRequiredBindings().size() == candidateAnchors.size() && ma.getRequiredBindings().containsAll(candidateAnchors)) {
				return true;
			}
		}
		return false;
	}

	private List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> computeAnchors(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc) {
		List<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> anchors =
				new ArrayList<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:pc.getDependencies()) {
			if(!isMergeableAlong(pn, pc)) {
				anchors.add(pn);
			}
		}
		return anchors;
	}

	private boolean isMergeableAlong(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn,
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc) {
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> dependant:pn.getDependantPatternConstraints()) {
			if(dependant != pc && (dependant.getDependencies().size() >= 2 || isFiltering(dependant))) {
				return false;
			}
		}
		return true;
	}

	private boolean filtersAnchor(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc,
			Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> anchors) {
		double maxDomainMappings = -1;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:anchors) {
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> domainMA = getDomainMatchingAction(pn);
			double mappings = getTotalMappings(domainMA);
			if(domainMA != null) {
				maxDomainMappings = Math.max(maxDomainMappings, mappings);
			}
		}
		
		if(maxDomainMappings == -1) {
			return true;
		}
		
		for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
			if(ma.getRequiredBindings().isEmpty() && getTotalMappings(ma) < maxDomainMappings) {
				return true;
			}
		}
		return false;
	}
	
	private MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> getDomainMatchingAction(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn) {
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:pn.getDependantPatternConstraints()) {
			if(pc.getConstraint() instanceof EClass) {
				for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
					if(ma instanceof MLSDMDomainMatchMatchingAction) {
						return ma;
					}
				}
			}
		}
		return null;
	}

	private double getTotalMappings(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMDomainMatchMatchingAction) {
			return metadataAdapter.getDomain((EClass) ma.getPatternConstraint().getConstraint()).size();
		}
		else if(ma instanceof MLSDMLinkLookupMatchingAction) {
			return metadataAdapter.getReferences((EReference) ((StoryPatternLink) ma.getPatternConstraint().getConstraint()).getFeature()).size();
		}
		else if(ma instanceof MLSDMIndexMatchingAction) {
			ICLConstraint constraint = ((ICLConstraint) ma.getPatternConstraint().getConstraint());
			return ((Index) variablesScope.getVariable(constraint.getIndexID()).getValue()).size();
		}
		else {
			return 0;
		}
	}

	private boolean isFiltering(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc) {
		CheckMatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> explicitCheckAction = pc.getExplicitCheckAction();
		if(explicitCheckAction instanceof MLSDMSPOExpressionCheckMatchingAction || explicitCheckAction instanceof MLSDMExpressionCheckMatchingAction) {
			int validInstances = metadataAdapter.getValidInstanceNumbers(explicitCheckAction.getPatternConstraint().getConstraint());
			if(validInstances != MLSDMMetadataIndex.EXPRESSION_NOT_REGISTERED) {
				return (double) validInstances < (double) metadataAdapter.getDomain((EClass) explicitCheckAction.getRequiredBindings().get(0).getSpo().getType()).size();
			}
			else if (explicitCheckAction.getPatternConstraint().getConstraint() instanceof ICLConstraint) {
				ICLConstraint constraint = (ICLConstraint) explicitCheckAction.getPatternConstraint().getConstraint();
				if(!indexFilters.containsKey(constraint)) {
					double domainSizes = 1;
					for(PatternNode<?, AbstractStoryPatternObject, ?, EClassifier, ?, ?> node:explicitCheckAction.getRequiredBindings()) {
						domainSizes *= metadataAdapter.getDomain((EClass) node.getSpo().getType()).size();
					}
					String indexId = constraint.getIndexID();
					Index index = (Index) variablesScope.getVariable(indexId).getValue();
					if(constraint.getOperation().getName().equals("CONTAINS")) {
						indexFilters.put(constraint, index.size() < domainSizes);
					}
					else if(constraint.getOperation().getName().equals("EXCLUDES")) {
						indexFilters.put(constraint, index.size() > 0);
					}
					else {
						indexFilters.put(constraint, true);
					}
				}
				return indexFilters.get(constraint);
			}
			else {
				return true;
			}
		}
		else if(explicitCheckAction instanceof MLSDMLinkCheckMatchingAction) {
			StoryPatternLink link = ((MLSDMLinkCheckMatchingAction) explicitCheckAction).getLink();
			double domainSize = ((double) metadataAdapter.getDomain((EClass) link.getSource().getType()).size() * (double) metadataAdapter.getDomain((EClass) link.getTarget().getType()).size());
			return metadataAdapter.getReferences((EReference) link.getFeature()).size() < domainSize;	//TODO polymorphism?
		}
		else {
			return false;		//TODO some other constraints obviously may filter
		}
	}

	@Override
	public int getIndex(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		return idMappings[node.getId()];
	}

	protected void addBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		if(index != -1) {
			currentBinding = currentBinding | (1 << index);
		}
	}
	
	protected void removeBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		if(index != -1) {
			currentBinding = currentBinding & (~(1 << index));
		}
	}
}
