package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.model.action.MLSDMIndexStagedMatchingAction;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.mlindices.Index;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.icl.iCL.ICLConstraint;
import de.mdelab.sdm.icl.iCL.ICLParameter;
import de.mdelab.sdm.icl.iCL.ICLVariable;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.MatchingAction;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMForwardCostCalculator
		extends
			MLSDMConstraintSensitiveCostCalculator {

	public MLSDMForwardCostCalculator(
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			MLSDMMetadataIndex metadataAdapter, CostModel costModel,
			VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super(searchModel, metadataAdapter, costModel, variablesScope);
	}

	@Override
	public Adornment[] calculateCotinuationCosts() {
		Adornment[] table = new Adornment[(int) Math.pow(2, searchModel.getPatternNodes().size())];
		Adornment initial = new Adornment();
		initial.totalCost = 1.0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:searchModel.getPatternNodes()) {
			initial.bindings[node.getId()] = node.isBound();
		}
		table[getBindingCode(initial)] = initial;
		
		@SuppressWarnings("unchecked")
		Collection<Adornment>[] configurations = (Collection<Adornment>[]) new Collection[searchModel.getPatternNodes().size() + 1];
		for(int i = 0; i < configurations.length; i++) {
			configurations[i] = new HashSet<Adornment>();
		}
		configurations[getBoundNodeNumber(initial)].add(initial);
		
		for(int i = getBoundNodeNumber(initial); i < configurations.length; i++) {
			Collection<Adornment> adornments = configurations[i];
			for(Adornment a:adornments) {
				Collection<Adornment> predecessors = computePredecessors(a);
				for(Adornment predecessor:predecessors) {
					int bindingCode = getBindingCode(predecessor);
					if(table[bindingCode] == null || table[bindingCode].totalCost > predecessor.totalCost) {
						int boundNodeNumber = getBoundNodeNumber(predecessor);
						configurations[boundNodeNumber].remove(table[bindingCode]);
						configurations[boundNodeNumber].add(predecessor);
						table[bindingCode] = predecessor;
					}
				}
			}
		}
		return table;
	}

	@Override
	protected Collection<Adornment> computePredecessors(Adornment a) {
		Collection<Adornment> predecessors = new ArrayList<Adornment>();
		for(PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc:a.unhandledConstraints) {
			for(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma:pc.getPossibleActions()) {
				/*
				 * skip inapplicable operations
				 */
				if(!isApplicable(ma, a.bindings)) {
					continue;
				}
				
				Adornment predecessor = getPredecessor(a, ma);
				if(predecessor != null) {
					predecessors.add(predecessor);
				}
			}
		}
		return predecessors;
	}

	@Override
	protected Adornment getPredecessor(Adornment a,
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		double cost = estimateCost(a, ma);
		
		if(cost != PatternConstraint.MATCHING_NOT_POSSIBLE) {
			boolean[] derivedBindings = new boolean[searchModel.getPatternNodes().size()];
			System.arraycopy(a.bindings, 0, derivedBindings, 0, derivedBindings.length);
			int newBindings = apply(ma, derivedBindings);

			return createDerivedAdornment(a, derivedBindings, cost, newBindings, ma);
		}
		
		return null;
	}
	
	private double estimateCost(Adornment a,
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		if(ma instanceof MLSDMIndexStagedMatchingAction) {
			ICLConstraint constraint = ((ICLConstraint) ma.getPatternConstraint().getConstraint());
			boolean foundUnbound = false;
			double domainSize = 1;
			double unmatchedDomainSize = 1;
			for(ICLParameter parameter:constraint.getParameters().getList()) {
				if(!(parameter instanceof ICLVariable)) {			//TODO handle other kinds of index parameters
					return PatternConstraint.MATCHING_NOT_POSSIBLE;
				}
				
				PatternNode<?, AbstractStoryPatternObject, ?, ?, ?, ?> node = nameMappings.get(((ICLVariable) parameter).getName());
				double partialDomainSize = metadataAdapter.getDomain((EClass) node.getSpo().getType()).size();
				domainSize *= partialDomainSize;
				
				if(foundUnbound || !a.bindings[node.getId()]) {
					unmatchedDomainSize *= partialDomainSize;
					foundUnbound = true;
				}
			}
			Index index = (Index) variablesScope.getVariable(constraint.getIndexID()).getValue();
			double filter = index.size() / domainSize;
			
			return unmatchedDomainSize * filter;
		}
		//TODO what about worst-case?
		else {
			return super.estimateCost(ma);
		}
	}

	private boolean isApplicable(
			MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma,
			boolean[] bindings) {
		for(PatternNode<?, ?, ?, ?, ?, ?> required:ma.getRequiredBindings()) {
			if(!bindings[required.getId()]) {
				return false;
			}
		}
		return true;
	}

	@Override
	protected Adornment createDerivedAdornment(Adornment a, boolean[] derivedBindings,
			double cost, int newBindings, MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma) {
		Collection<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> uncheckedConstraints = new ArrayList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.unhandledConstraints);
		LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> matchingOrder = new LinkedList<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.matchingOrder);
		matchingOrder.add(ma.getPatternConstraint());
		LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> actions = new LinkedList<MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>(a.actions);
		actions.add(ma);
		double filtering = 1;
		for(Iterator<PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> it = uncheckedConstraints.iterator(); it.hasNext();) {
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pc2 = it.next();
			if(isChecked(pc2, derivedBindings)) {
				it.remove();
				if(pc2 != ma.getPatternConstraint()) {
					matchingOrder.add(pc2);
					actions.add(pc2.getExplicitCheckAction());
					filtering *= getFiltering(pc2.getExplicitCheckAction());
				}
			}
		}
		
		return new Adornment(derivedBindings, uncheckedConstraints, a.totalCost + (a.front * cost * newBindings), matchingOrder, actions, 1.0, a.front * cost * filtering);	//TODO applied filtering relevant here?
	}

	@Override
	protected int apply(MatchingAction<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> ma, boolean[] bindings) {
		int newBindings = 0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> pn:ma.getPatternConstraint().getDependencies()) {
			int index = pn.getId();
			if(!bindings[index]) {
				newBindings++;
			}
			bindings[index] = true;
		}
		return newBindings;
	}
	
}
