package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.Adornment;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternConstraint;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.SearchModel;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMHedgingCostStrategy extends MLSDMHybridCostStrategy{

	protected MLSDMContinuationCostCalculator pessimisticCalculator;
	protected Adornment[] pessimisticTable;
	protected double tolerance;
	
	public MLSDMHedgingCostStrategy(
			MLSDMHybridSelectionStrategy selectionStrategy,
			SearchModel<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> searchModel,
			CostModel costModel,
			VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope,
			MLSDMMetadataIndex metadataAdapter,
			double tolerance) {
		super(selectionStrategy, searchModel, costModel, variablesScope,
				metadataAdapter);
		this.tolerance = tolerance;
	}

	@Override
	public void initialize() {
		super.initialize();
		pessimisticCalculator = new MLSDMContinuationCostCalculator(searchModel, metadataAdapter, CostModel.WORST_CASE, variablesScope);
		pessimisticTable = pessimisticCalculator.calculateCotinuationCosts();
	}
	
	@Override
	public double calculateMatchingCost(
			PatternConstraint<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> patternConstraint) {
		if(patternConstraint.getCardinality() == PatternConstraint.CHECK_ONLY) {
			return PatternConstraint.CHECK_ONLY;
		}
		
		int oldBindingCode = selectionStrategy.getCurrentBinding();
		int newBindingCode = oldBindingCode;
		int newBindings = 0;
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:patternConstraint.getDependencies()) {
			if(!node.isBound()) {
				newBindingCode = newBindingCode | (1 << selectionStrategy.getIndex(node));
				newBindings++;
			}
		}
		
		double constraintCardinality = patternConstraint.getCardinality();
		double oldFiltering = continuationTable[oldBindingCode].appliedFilters;
		double newFiltering = continuationTable[newBindingCode].appliedFilters;
		double constraintFiltering = costCalculator.getFiltering(patternConstraint.getExplicitCheckAction());
		double filtering = (constraintFiltering == 0 || oldFiltering == 0) ? 0 : newFiltering / (constraintFiltering * oldFiltering);
		double expectedCost = 	(constraintCardinality * newBindings) +
									(constraintCardinality * filtering * continuationTable[newBindingCode].totalCost);
		
		double worstCost = constraintCardinality * newBindings + constraintCardinality * pessimisticTable[newBindingCode].totalCost;
		
		tolerance = searchModel.getPatternNodes().size() - searchModel.getBoundPatternNodeNumber();	//TODO
		
		return worstCost <= tolerance * pessimisticTable[oldBindingCode].totalCost ? expectedCost : PatternConstraint.MATCHING_NOT_POSSIBLE;
	}

}
