package de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid;

import java.util.Collection;
import java.util.LinkedList;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.MLSDMMetadataIndex;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMCheapestFirstSelectionStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.MLSDMMatchingCostStrategy;
import de.mdelab.mlsdm.interpreter.searchModel.patternMatcher.strategy.hybrid.MLSDMContinuationCostCalculator.CostModel;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.patternmatcher.searchModelBased.PatternNode;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

public class MLSDMHybridSelectionStrategy
		extends
			MLSDMCheapestFirstSelectionStrategy {

	protected MLSDMMetadataIndex metadataAdapter;
	protected CostModel costModel;
	protected VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope;
	
	protected int currentBinding;
	protected LinkedList<Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>> bindingsStack;
	
	public MLSDMHybridSelectionStrategy(MLSDMMetadataIndex metadataAdapter, CostModel costModel,
										VariablesScope<?, ?, ?, ?, ?, ?, ?, ?, MLExpression> variablesScope) {
		super();
		this.metadataAdapter = metadataAdapter;
		this.costModel = costModel;
		this.variablesScope = variablesScope;
		this.currentBinding = 0;
		this.bindingsStack = new LinkedList<Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>>>();
	}
	
	@Override
	protected MLSDMMatchingCostStrategy createMatchingCostStrategy() {
		return new MLSDMHybridCostStrategy(this, searchModel, costModel, variablesScope, metadataAdapter);
	}

	@Override
	protected void doUpdate(Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> affectedNodes) {
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:affectedNodes) {
			addBinding(node);
		}
		bindingsStack.push(affectedNodes);
	}

	@Override
	protected void doRollBackLastUpdate() {
		Collection<PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>> affectedNodes = bindingsStack.pop();
		for(PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node:affectedNodes) {
			removeBinding(node);
		}
	}
	
	protected void addBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		currentBinding = currentBinding | (1 << index);
	}
	
	protected void removeBinding(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		int index = getIndex(node);
		currentBinding = currentBinding & (~(1 << index));
	}

	public int getIndex(
			PatternNode<StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> node) {
		return node.getId();
	}

	public int getCurrentBinding() {
		return currentBinding;
	}
	
}
